#!/usr/bin/env python
# coding: utf-8
# Copyright (C) 2017-2019 Matthieu Ancellin
# See LICENSE file at <https://github.com/mancellin/capytaine>

from capytaine.bodies.predefined.spheres import Sphere
from capytaine.bodies.predefined.cylinders import HorizontalCylinder, VerticalCylinder
from capytaine.bodies.predefined.rectangles import Rectangle, RectangularParallelepiped
